try:
    from shapely.geometry import Polygon,MultiPolygon, LineString, Point
    from shapely.ops import split as shape_split
except:
    print('ERROR: Some functions in help_functions need Shapely package. But it could not be imported. Behaviour can be unpredictable!')
from typing import List, Tuple
from rhdhv_fem.shapes.shapes import Shapes
import rhdhv_fem as fem
from rhdhv_fem import fem_math as fm
import viiapackage as viia
import numpy as np
import itertools as itr
def plot(data, transpose=False):
    import matplotlib
    import matplotlib.pyplot as plt
    matplotlib.use(Config.MPL_GUI())
    data = np.array(data)    
    for dat in data:
        if not transpose:
            plt.plot(*dat)
        else:
            plt.plot(*dat.T)
    plt.show()
    
    
def choose_master_slave(project, shape1, shape2):
    # Chooseing master and slave from 2 shapes. Currently based on how many 
    # points of one shape are contained by another. If one shape has less points contained in another
    # then that shape is master.
    points1 = shape1.contour.get_points() 
    points2 = shape2.contour.get_points() 
    count1 = 0
    for point in points1:
        if fm.fem_point_in_surface(point, points2):
            count1 +=1
    count2 = 0
    
    for point in points2:
        if fm.fem_point_in_surface(point, points1):
            count2 +=1
    if count1> count2:
        return shape2, shape1 # master, slave
    else:
        return shape1, shape2 # master, slave
        
def convert_to_plygon(shape):
    points = np.array(shape.contour.get_points())
    points_2d = points[:,:-1]
    polygon = Polygon(points_2d)    
    return polygon, np.average(points[:,-1])
  
def adjust_shapes(project: 'Project',shapes: list, precision: float, simplify: float):
    """
    Adjust horizontal flat 3d shapes so they do not overlap.
    Input:
        project {Project} -- Project object
        shapes {List} -- List of shapes to adjust
        precission {float} -- precission to consider shapes in one level.
        simplify {float} -- simplification parameter to adust resulting geometry.
    """
    #TODO: Allow any oriantion as long as they are in one plane
    #TODO: Rewrite with VIIA functions
    # Find intersections between shape lines
    for shape1, shape2 in itr.combinations(shapes,2):
        # TODO: Adust master slave relationship and select governing strip by better means
        master, slave = choose_master_slave(project,shape1, shape2)
        slave_shape, z_level_s = convert_to_plygon(slave)
        master_shape, z_level_m = convert_to_plygon(master)
        
        # determine if two shapes are in one level
        if abs(z_level_m - z_level_s) < 10**(-precision):
            #determine if master intersects slave
            if master_shape.overlaps(slave_shape):
                slave_shape_new = slave_shape.difference(master_shape)
                # change the slave
                try:
                    coords = np.array(slave_shape_new.simplify(simplify).buffer(0).exterior.coords)
                    _, idx = np.unique(coords,axis=0,return_index=True)
                    new_contour = coords[np.sort(idx)]
                except AttributeError:                    
                    #sometimes cuts produce invalid geometry a valid should be selected:
                    areas = [x.area for x in slave_shape_new.geoms]
                    #selection is based on max area
                    coords = np.array(slave_shape_new.geoms[areas.index(max(areas))].simplify(simplify).buffer(0).exterior.coords)
                    _, idx = np.unique(coords,axis=0,return_index=True)
                    new_contour = coords[np.sort(idx)]        
                except ValueError:
                    plot([slave_shape_new.exterior.coords], transpose=True)
                new_contour = np.c_[new_contour, np.full(len(new_contour),z_level_s)]
                try:
                    slave.contour = project.create_polyline(new_contour.tolist())
                except Exception as e:
                    print(e)
                    print(new_contour.tolist())
                    plot([new_contour], transpose=True)


def create_simple_wall(project:"ViiaProject", name: str, material: str, thickness: int, starting_point: List[float], length: List[float], height: float) -> Tuple[Shapes, List[float]]:
    """
    Creates a wall from simple parameters
    
    Input:
        project {Project} -- Project object
        name {str} -- Name of the wall
        material {str} -- Name of the material
        thickness {int} -- Wall thickness in mm
        starting_point {list} -- Starting point of the wall
        length {list} -- 2d vector of wall length
        height {float} -- height of the wall

    Output:
        wall (Shapes) -- Newly created wall
        end_point (list) -- 3d coordinates of the endpoint of the wall
        
    """
    wall: Shapes = project.viia_create_wall(name, material, thickness,
                                            [[[starting_point[0], starting_point[1], starting_point[2]],
                                              [starting_point[0]+length[0],
                                               starting_point[1]+length[1], starting_point[2]],
                                              [starting_point[0]+length[0], starting_point[1] +
                                               length[1], starting_point[2]+height],
                                                [starting_point[0], starting_point[1], starting_point[2]+height]]])
    end_point = [starting_point[0]+length[0],
                 starting_point[1]+length[1], starting_point[2]]

    return wall, end_point

def rot(vector,theta):
    """
    Rotates 2-D vector around z-axis
    
    Input:
        axis (list): List of 2d coordinates that describe axis to be rotated.
        theta (float): angle in degrees to rotate around specified axis
    
    Output:
        matrix: rotated vector
    """
    theta=theta/180.0*np.pi    
    R = np.array([[np.cos(theta), -np.sin(theta)],[np.sin(theta), np.cos(theta)]])
    return np.dot(R,vector)

def rotation_matrix(axis, theta):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    
    Input:
        axis (list): List of 3d coordinates that describe axis around which a vector needs to be rotated.
        theta (float): angle in dgerees to rotate around specified axis
    
    Output:
        matrix: rotational matrix
    
    """
    theta=theta/180.0*np.pi    
    axis = np.asarray(axis)
    axis = axis / np.sqrt(np.dot(axis, axis))
    a = np.cos(theta / 2.0)
    b, c, d = -axis * np.sin(theta / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                     [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                     [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])